/*************************************************************************/
/*                                                                	 */
/*	Evaluation of a test on a continuous valued attribute	  	 */
/*	-----------------------------------------------------	  	 */
/*								  	 */
/*************************************************************************/

#include "buildex.i"

float	BETA;

extern 
ItemNo  *Inf,           /* Inf[a] = min index for values of attribute a */ 
  *Sup,           /* Sup[a] = max index for values of attribute a */ 
  *PBestI,        /* PBestI[a] = min index for values if attribute a is selected */ 
  *PBestS,        /* PBestS[a] = max index for values if attribute a is selected */ 
  *PBestL,        /* PBestL[a] = min index for values above threshold 
		     if attribute a is selected */ 
  *PBestH,        /* PBestH[a] = max index for values below threshold 
		     if attribute a is selected */ 
  MaxDistinctValues;

ItemNo *NDistinctValues, /* NDistinctValues[a] = number of distinct values for attribute a */
  AVC_ROW = 0, /* NDistinctValues[a] = 0 iff values in a have not yet been sorted */
  AVC_MAX_ROW; 
			       
float **DistinctValues; /* DistinctValues[a][i] = i-th distinct value of a, for i >= 1 
			   DistinctValues[a][0] = the Unknown value */

float  **AVC; 

/*************************************************************************/
/*								  	 */
/*  Continuous attributes are treated as if they have possible values	 */
/*	0 (unknown), 1 (less than cut), 2(greater than cut)	  	 */
/*  This routine finds the best cut for items Fp through Lp and sets	 */
/*  Info[], Gain[] and Bar[]						 */
/*								  	 */
/*************************************************************************/


    EvalContinuousAtt(Att, Fp, Lp)
/*  -----------------  */ 
    Attribute Att;
    ItemNo Fp, Lp; 
{ 
    ItemNo i, j, k, BestI, BestII, Xp, Tries=0, MaxI, MinI, NVal;
    register ItemNo pos;
    ItemCount Items, KnownItems, LowItems, PrevLI, MinSplit, CountItems();
    ClassNo c;
    Boolean AttSorted;
    float SplitG, SplitI, BestVal, BestInfo, BaseInfo, ThreshCost,
      ComputeGainContin(), TotalInfo(), Worth(); 
    void Swap(), SwapI(), SetValues();
    int  SetValuesInteger();

    Verbosity(2) printf("\tAtt %s", AttName[Att]);
    Verbosity(3) printf("\n");

    ResetFreq(2);
    
    /*  Omit and count unknown values */
    
    Items = CountItems(Fp, Lp);
 
    if ( !NDistinctValues[Att])
      {
	/* Attribute Att values not yet sorted. We are in the root! */
 
	Xp = Fp;
	ForEach(i, Fp, Lp)
	  {
	    /* Use CPVal to refer to continuous values */
	    if ( CPVal(Item[i],Att) == Unknown )
	      {
		Freq[ 0 ][ Class(Item[i]) ] += Weight[i];
		SwapI(Xp, i);
		Xp++;
	      }
	  }
	
	ValFreq[0] = 0;
	ForEach(c, 0, MaxClass)
	  {
	    ValFreq[0] += Freq[0][c];
	  }
	
	KnownItems = Items - ValFreq[0];
	UnknownRate[Att] = 1.0 - KnownItems / Items;
	
	/*  Special case when very few known values  */
	
	if ( KnownItems < 2 * MINOBJS )
	  {
	    Verbosity(2) printf("\tinsufficient cases with known values\n");
	    
	    Gain[Att] = -Epsilon;
	    Info[Att] = 0.0;
	    return;
	  }

	/* Use Quicksort on Continuous Values */
	
	if (SpecialStatus[Att] == INTEGER) 
	  {
	    if( SetValuesInteger(Att, Xp, Lp) )
	      {
		AttSorted = 0;
		NVal = Sup[Att] - Inf[Att] + 1;
	      }
	    else 
	      SpecialStatus[Att] = Nil;
	  }
	
	if (SpecialStatus[Att] != INTEGER)
	  {
	    QuicksortCP(Xp, Lp, Att, Swap);
	    
	    /* Initialise DistinctValues[Att] */
	    
	    SetValues(Att, Xp, Lp);
	    NVal = Sup[Att] - Inf[Att] + 1;
	    if( (NVal <= AVC_ROW) && ( NVal/(Lp - Xp + 1) <= BETA ) )
	      AttSorted = 0;
	    else
	      AttSorted = 1;
	  }
      }
    else
      {
	/* Attribute Att values already sorted! */
	
	NVal = Sup[Att] - Inf[Att] + 1;
	Xp = Fp;
	ForEach(i, Fp, Lp)
	  {
	    /* Use CVal to refer to continuous values */
	    if ( CVal(Item[i],Att) == Unknown )
	      {
		Freq[ 0 ][ Class(Item[i]) ] += Weight[i];
		Swap(Xp, i);
		Xp++;
	      }
	  }
	
	ValFreq[0] = 0;
	ForEach(c, 0, MaxClass)
	  {
	    ValFreq[0] += Freq[0][c];
	  }

	KnownItems = Items - ValFreq[0];
	UnknownRate[Att] = 1.0 - KnownItems / Items;

	/*  Special case when very few known values  */

	if ( KnownItems < 2 * MINOBJS )
	  {
	    Verbosity(2) printf("\tinsufficient cases with known values\n");

	    Gain[Att] = -Epsilon;
	    Info[Att] = 0.0;
	    return;
	  }

	if( (NVal <= AVC_ROW) && ( NVal/(Lp - Xp + 1) <= BETA ) )
	  {
	    AttSorted = 0;
	  }
	else
	  {
	    AttSorted = 1;
	    CountingSort(Xp, Lp, Att, Swap);
	  }
      }
    
    /*  Count base values and determine base information  */
    
 
    if( AttSorted )
      { 

	/* Strategy one or two */
	
	ForEach(i, Xp, Lp)
	  {
	    Freq[ 2 ][ Class(Item[i]) ] += Weight[i];
	  }
	
	BestVal = 0;
	BestI = None;
	BestInfo = 0;
	
	BaseInfo = TotalInfo(Freq[2], 0, MaxClass) / KnownItems;
	
	/*  Try possible cuts between items i and i+1, and determine the
	    information and gain of the split in each case.  We have to be wary
	    of splitting a small number of items off one end, as we can always
	    split off a single item, but this has little predictive power.  */
	
	MinSplit = 0.10 * KnownItems / (MaxClass + 1);
	if ( MinSplit <= MINOBJS ) MinSplit = MINOBJS;
	else
	  if ( MinSplit > 25 ) MinSplit = 25;
	
	LowItems = 0;
	ForEach(i, Xp, Lp - 1)
	  {
	    c = Class(Item[i]);
	    LowItems   += Weight[i];
	    Freq[1][c] += Weight[i];
	    Freq[2][c] -= Weight[i];
	    
	    if ( LowItems < MinSplit ) continue;
	    else
	      if ( LowItems > KnownItems - MinSplit ) break;
	    
	    if ( CVal(Item[i],Att) < CVal(Item[i+1],Att) - 1E-5 )
	      {
		ValFreq[1] = LowItems;
		ValFreq[2] = KnownItems - LowItems;
		SplitG = ComputeGainContin(BaseInfo, UnknownRate[Att], 2, KnownItems);
		SplitI = TotalInfo(ValFreq, 0, 2);
				
		if( SplitG > BestVal )
		  {
		    BestI = i;
		    BestVal = SplitG;
		    BestInfo = SplitI;
		  }
		
		Tries++;
				
		Verbosity(3)
		  {	printf("\t\tCut at %.3f  (gain %.3f, val %.3f):",
			       ( CVal(Item[i],Att) + CVal(Item[i+1],Att) ) / 2,
			       SplitG,
			       Worth(SplitI / Items, SplitG, Epsilon));
		  PrintDistribution(Att, 2, true);
		  }
	      }
	  }
	
	/*  Find the best attribute according to the given criterion.
	    The piece of code of C4.5 has been moved into the previuous
	    while statement.
	*/
	
	ThreshCost = Log(Tries) / Items;
	
	/*  If a test on the attribute is able to make a gain,
	    set the best break point, gain and information  */ 
	
	if ( BestVal <= ThreshCost )
	  {
	    Gain[Att] = -Epsilon;
	    Info[Att] = 0.0;
	    
	    Verbosity(2) printf("\tno gain\n");
	  }
	else
	  {
	    BestVal -= ThreshCost;
	    
	    /* Set PBest?[Att] in case Att will be selected */
	    
	    PBestI[Att] = PVal(Item[BestI],Att);
	    PBestS[Att] = PVal(Item[BestI+1],Att);
	    
	    PBestL[Att] = PVal(Item[Xp],Att);
	    PBestH[Att] = PVal(Item[Lp],Att);
	    
	    
	    Bar[Att]  = (CVal(Item[BestI],Att) + CVal(Item[BestI+1],Att)) / 2;
	    Gain[Att] = BestVal;
	    Info[Att] = BestInfo / Items;
	    
	    Verbosity(2)
	      printf("\tcut=%.3f, inf %.3f, gain %.3f\n",
		     Bar[Att], Info[Att], Gain[Att]);
	  }
      }
    else
      {
	/* Strategy three */

	ForEach(c, 0, MaxClass)
	  {
	    ForEach(i, 0, Sup[Att] - Inf[Att])
	      {
		AVC[c][i] = 0;
	      }
	  }
	

	/* MinI (MaxI) is the minimum (maximum) index */

	MinI = MaxI = PVal(Item[Xp], Att) - Inf[Att];
       
	ForEach(i, Xp, Lp)
	  {
	    Freq[ 2 ][ Class(Item[i]) ] += Weight[i];
	    pos = PVal(Item[i], Att) - Inf[Att];
	    AVC[ Class(Item[i]) ][ pos ] += Weight[i];
	    
	    if (pos > MaxI) MaxI = pos;
	    if (pos < MinI) MinI = pos;
	  }
       
	BestVal = 0;
	BestI = None;
	BestInfo = 0;
	
	BaseInfo = TotalInfo(Freq[2], 0, MaxClass) / KnownItems;
	
	/*  Try possible cuts between items i and i+1, and determine the
	    information and gain of the split in each case.  We have to be wary
	    of splitting a small number of items off one end, as we can always
	    split off a single item, but this has little predictive power.  */
	
	MinSplit = 0.10 * KnownItems / (MaxClass + 1);
	if ( MinSplit <= MINOBJS ) MinSplit = MINOBJS;
	else
	  if ( MinSplit > 25 ) MinSplit = 25;
	
	LowItems = 0;
	
	j = MinI; /* j store previuos index with Weight > 0 */

	ForEach(i, MinI, MaxI-1)
	  {
	    PrevLI = LowItems ;
	    ForEach(c, 0, MaxClass)
	      {
		LowItems += AVC[c][i];
		Freq[1][c] += AVC[c][i];
		Freq[2][c] -= AVC[c][i];
	      }
	    
	    if ( PrevLI == LowItems ) continue; 
	    else
	      {
		AVC[0][j] = i; /* AVC[0][previous index] = i */
		j = i;         /* it is used to find next index */
	      }
	   
	    if ( LowItems < MinSplit ) continue;
	    else
	      if ( LowItems > KnownItems - MinSplit ) 
		  break;
	    
	    if ( !(DistinctValues[Att][i + Inf[Att]] < DistinctValues[Att][i + 1 + Inf[Att]] - 1E-5 ) )
	      {
		/* The next distinct value is not greater by at least 1E-5 
		   However, it may be the case that the next distinct value
		   is not present in the cases of the node. So, to be sure
		   that the iteration has to be interrupted, we have to find
		   the next distinct value with non-zero weight */

		k = i;
		do {
		  k++;
		  PrevLI = 0;
		  ForEach(c, 0, MaxClass)
		    PrevLI += AVC[c][k];
		} while( PrevLI == 0 );
		
		if( !(DistinctValues[Att][i + Inf[Att]] < DistinctValues[Att][k + Inf[Att]] - 1E-5 ) )
		  continue;
	      }
	    
	    ValFreq[1] = LowItems;
	    ValFreq[2] = KnownItems - LowItems;
	    SplitG = ComputeGainContin(BaseInfo, UnknownRate[Att], 2, KnownItems);
	    SplitI = TotalInfo(ValFreq, 0, 2);
	   	   
	    if( SplitG > BestVal)
	      {
		BestI = i;
		BestVal = SplitG;
		BestInfo = SplitI;
	      }
	    
	    Tries++;
	    
	    Verbosity(3)
	      {	 printf("\t\tCut at %.3f  (gain %.3f, val %.3f):",
			/* This value is an approximation. The correct value is computed later
			   for best split */
			( DistinctValues[Att][i + Inf[Att]] + DistinctValues[Att][k + Inf[Att]] ) / 2,
			SplitG,
			Worth(SplitI / Items, SplitG, Epsilon));
	      PrintDistribution(Att, 2, true);
	      }
	  }

	AVC[0][j] = MaxI;
	ThreshCost = Log(Tries) / Items;
       
	if ( BestVal <= ThreshCost )
	  {
	    Gain[Att] = -Epsilon;
	    Info[Att] = 0.0;
	    
	    Verbosity(2) 
	      printf("\tno gain\n");
	  }
	else
	  {
	    BestVal -= ThreshCost;
	    
	    BestII = AVC[0][BestI];
	    
	    BestII += Inf[Att];
	    BestI += Inf[Att];
	    
	    /* Set PBest?[Att] in case Att will be selected */
	    
	    PBestI[Att] = BestI;
	    PBestS[Att] = BestII;
	    
	    PBestL[Att] = MinI + Inf[Att];
	    PBestH[Att] = MaxI + Inf[Att];
	    	   
	    Bar[Att]  = (DistinctValues[Att][BestI] + DistinctValues[Att][BestII])
	     / 2;
	    Gain[Att] = BestVal;
	    Info[Att] = BestInfo / Items;
	    
	    Verbosity(2) 
	      printf("\tcut=%.3f, inf %.3f, gain %.3f\n",
		     Bar[Att], Info[Att], Gain[Att]);
	  }
      }
}


/*************************************************************************/
/*								  	 */
/*  Adjust the dimension of AVC at the root after distinct values of 	 */
/*  an attribute have been computed                         	  	 */
/*								  	 */
/*************************************************************************/

void UpdateND(Att)
/*  -----------------  */ 
    Attribute Att;
{
  ClassNo cl;
  ItemNo NewDim_AVC;
    
  if( NDistinctValues[Att] > MaxDistinctValues )
    {
      MaxDistinctValues = NDistinctValues[Att]; 
      if( (EC_STRATEGY == 3) && AVC_ROW)
	{
	  NewDim_AVC = Min(MaxDistinctValues,AVC_MAX_ROW);
	  if( NewDim_AVC > AVC_ROW )
	    {
	      AVC_ROW = NewDim_AVC;
	      ForEach(cl, 0, MaxClass)
		{
		  free(AVC[cl]);
		  AVC[cl] = (float *) calloc(AVC_ROW, sizeof(float));
		}
	    }
	}
      else if (EC_STRATEGY == 3)
	{
	  AVC_ROW = Min(MaxDistinctValues,AVC_MAX_ROW);
	  AVC = (float **) calloc(MaxClass + 1, sizeof(float*));
	  ForEach(cl, 0, MaxClass)
	    AVC[cl] = (float *) calloc(AVC_ROW, sizeof(float));
	}
      
    }
}


/*************************************************************************/
/*                                                                	 */
/*  Initialise the DistinctValues[Att] array for attribute Att         	 */
/*                                                                	 */
/*************************************************************************/

void    SetValues(Att, Fp, Lp)
/*  -----------------  */ 
    Attribute Att;
    ItemNo Fp, Lp; 
{ 
  ItemNo i,c;
  float  v,n;
  
  /* Count in c the distinct values in Fp..Lp */
  
  v = CPVal(Item[Fp],Att);
  c = 1;
  
  ForEach(i, Fp+1, Lp)
    if ( v != (n = CPVal(Item[i],Att)) )
      {
	c++ ;
	v = n ;
      }
  
  /* Allocate memory */
  
  DistinctValues[Att] = (float *) calloc(c+1, sizeof(float));
  
  /* Second pass to set DistinctValues[Att][c] as the c-th
     distinct values. Also, Item[i] is set to have Att value
     as the index corresponding to its actual continuous value.
     
     Note that array is ordered, since SetValues is called after 
     QuickSort!
  */
  
  c = 0;
  DistinctValues[Att][c] = Unknown; /* 0-th value is Unknown */
  
  ForEach(i, 0, Fp-1)
    PVal(Item[i],Att) = c;
  
  c++;
  v = DistinctValues[Att][c] = CPVal(Item[Fp],Att);
  PVal(Item[Fp],Att) = c;
  
  ForEach(i, Fp+1, Lp)
    if ( v != (n = CPVal(Item[i],Att)) )
      {
	c++ ;
	v = DistinctValues[Att][c] = n ;
	PVal(Item[i],Att) = c;
      }
    else
      {
	PVal(Item[i],Att) = c;
      }
  
  /* Initialise Inf and Sup and NDistinctValues */
  Inf[Att] = 1;
  Sup[Att] = 
    NDistinctValues[Att] = c;
  
  UpdateND(Att);
}



/*************************************************************************/
/*                                                                	 */
/*  Initialise the DistinctValues[Att] array for attribute Att         	 */
/*  specialised for attributes with only integer values                	 */
/*                                                                	 */
/*************************************************************************/

int  SetValuesInteger(Att, Fp, Lp)
/*  -----------------  */ 
    Attribute Att;
    ItemNo Fp, Lp; 
{ 
  ItemNo i,j,maxrange;
  float  v;
  int	max, min;
  void Swap();

  /* Count in c the distinct values in Fp..Lp */
  
  min = max = (int) CPVal(Item[Fp],Att);
  
  i = Fp + 1 ;

  /* This algorithm may waste some memory in
     the DistinctValues array. So, we restrict
     its applicability 
  */

  maxrange = Min(256,AVC_MAX_ROW); 
  
  while( (i <= Lp) && (max - min) < maxrange)
    {
      v = CPVal(Item[i++],Att);
      
      if( v < min ) min = v;
      if( v > max ) max = v;
    }

  if(  (max - min) < maxrange )
    {
      DistinctValues[Att] = (float *) calloc((max - min + 2), sizeof(float));
      
      DistinctValues[Att][0] = Unknown;

      ForEach(i, 0, Fp-1)
	PVal(Item[i],Att) = 0;
      
      /* Counting of integers that actually are in
	 the training set */

      ForEach(i, 1, max - min + 1)
	DistinctValues[Att][i] = 0;
   
      ForEach(i, Fp, Lp)
	{
	  j = (CPVal(Item[i],Att) - min + 1);
	  DistinctValues[Att][j] = 1;
	}
      
      DistinctValues[Att][1] += min - 1;
      ForEach(i, 2, max - min + 1)
	{
	  DistinctValues[Att][i] += DistinctValues[Att][i-1];
	}
      
      ForEach(i, Fp, Lp)
	{
	  j = (CPVal(Item[i],Att) - min + 1);
	  PVal(Item[i],Att) = (int) (DistinctValues[Att][j]) - min + 1;
	}
      
      j = min-1; Fp = 0;
      ForEach(i, 1, max - min + 1)
	if( DistinctValues[Att][i] != j )
	  {
	    DistinctValues[Att][++Fp] = j = DistinctValues[Att][i];
	  }
      
      Inf[Att] = 1;
      Sup[Att] = 
	NDistinctValues[Att] = Fp;
      
      UpdateND(Att);
      return 1;
    }
  else
    {
      return 0;
    }
}


/*************************************************************************/
/*                                                                	 */
/*  Change a leaf into a test on a continuous attribute           	 */
/*                                                                	 */
/*************************************************************************/


ItemNo   ContinTest(Node, Att)
/*  ----------  */
    Tree Node;
    Attribute Att;
{
    float Thresh, GreatestValueBelow();

    Sprout(Node, 2);

    Thresh = GreatestValueBelow(Att, Bar[Att]); 

    Node->NodeType	= ThreshContin;
    Node->Tested	= Att;
    Node->Cut		=
    Node->Lower		=
    Node->Upper		= Thresh;
    Node->Errors        = 0;

}


/*************************************************************************/
/*                                                                	 */
/*  Return the greatest value of attribute Att below threshold t  	 */
/*  Implements a binary search on DistinctValues[Att]                    */
/*                                                                	 */
/*************************************************************************/


float GreatestValueBelow(Att, t)
/*    ------------------  */
    Attribute Att;
    float t;
{
    register ItemNo i, j, m;
    float *Base;

    i = PBestI[Att];     /* Lower Bound */
    j = PBestS[Att] + 1; /* Upper Bound */
    Base = DistinctValues[Att];
    
    while( (i+1) != j)
      {
	m = (i + j) / 2;
	if( Base[m] <= t )
	  i = m;
	else
	  j = m;
      }
    
    if( t < Base[m])
      return Base[m-1];
    else
      return Base[m];
}



/*************************************************************************/
/*                                                                	 */
/*  In the TRIALS option, this procedure performs the initial sorting	 */
/*  and setting of DistinctValues and NDistinctValues data structues	 */
/*                                                                	 */
/*************************************************************************/


 SortAtTheBeginning()
{ 
  Attribute Att;
  ItemNo Xp,Fp,Lp,i;
  void SwapI();
  
  
  ForEach(Att, 0, MaxAtt) 
    if ( !MaxAttVal[Att] )
      { 
	/*  continuous attribute  */
	
	Xp = Fp = 0;
	Lp = MaxItem;
	ForEach(i, Fp, Lp)
	  if ( CPVal(Item[i],Att) == Unknown )
	    {
	      SwapI(Xp, i);
	      Xp++;
	    }
	
	if (SpecialStatus[Att] == INTEGER) 
	  {
	    if( !SetValuesInteger(Att, Xp, Lp) )
	      SpecialStatus[Att] = Nil;
	  }
	
	if (SpecialStatus[Att] != INTEGER)
	  {
	    QuicksortCP(Xp, Lp, Att, SwapI);
	    SetValues(Att, Xp, Lp);
	  }
      }
}

